"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Generator = void 0;
const random_js_1 = require("random-js");
const colors_1 = __importDefault(require("colors"));
class Generator {
    constructor(dbConnector, schema, callback = () => null) {
        this.dbConnector = dbConnector;
        this.schema = schema;
        this.callback = callback;
        this.continue = false;
        if (schema.settings.seed) {
            this.random = new random_js_1.Random(random_js_1.MersenneTwister19937.seed(schema.settings.seed));
        }
        else {
            this.random = new random_js_1.Random(random_js_1.MersenneTwister19937.autoSeed());
        }
    }
    gotoNextTable() {
        this.continue = true;
    }
    empty(table) {
        return __awaiter(this, void 0, void 0, function* () {
            this.callback({ currentTable: table.name, step: 'empty', state: 'RUNNING', currentValue: 0, max: 1 });
            yield this.dbConnector.emptyTable(table);
            this.callback({ currentTable: table.name, step: 'empty', state: 'DONE', currentValue: 1, max: 1 });
        });
    }
    getForeignKeyValues(table, tableForeignKeyValues = {}, runRows) {
        return __awaiter(this, void 0, void 0, function* () {
            const columns = table.columns.filter((column) => column.foreignKey);
            const foreignKeyPromises = columns.map((column, index) => {
                this.callback({ currentTable: table.name, step: 'foreignkeys', state: 'RUNNING', currentValue: index, max: columns.length });
                const foreignKey = column.foreignKey;
                return this.dbConnector.getValuesForForeignKeys(table.name, column.name, foreignKey.table, foreignKey.column, runRows, column.unique, foreignKey.where).then((values) => {
                    tableForeignKeyValues[`${column.name}_${foreignKey.table}_${foreignKey.column}`] = values;
                });
            });
            yield Promise.all(foreignKeyPromises);
            this.callback({ currentTable: table.name, step: 'foreignkeys', state: 'DONE', currentValue: columns.length, max: columns.length });
        });
    }
    fillTables(reset = false) {
        return __awaiter(this, void 0, void 0, function* () {
            this.beforeAll();
            for (const table of this.schema.tables) {
                yield this.fill(table, reset);
            }
            this.afterAll();
        });
    }
    fill(table, reset = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (reset)
                yield this.empty(table);
            let handleTriggers = table.disableTriggers || (table.disableTriggers === undefined && this.schema.settings.disableTriggers);
            if (handleTriggers)
                yield this.dbConnector.disableTriggers(table.name);
            yield this.before(table);
            const insertedRows = yield this.generateData(table);
            if (insertedRows)
                yield this.after(table);
            if (handleTriggers)
                yield this.dbConnector.enableTriggers(table.name);
        });
    }
    beforeAll() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.schema.settings.beforeAll.length === 0)
                return;
            for (let i = 0; i < this.schema.settings.beforeAll.length; i++) {
                this.callback({ currentTable: '', step: 'beforeAll', state: 'RUNNING', currentValue: i, max: this.schema.settings.beforeAll.length });
                yield this.dbConnector.executeRawQuery(this.schema.settings.beforeAll[i]);
            }
            this.callback({ currentTable: '', step: 'beforeAll', state: 'DONE', currentValue: this.schema.settings.beforeAll.length, max: this.schema.settings.beforeAll.length });
        });
    }
    afterAll() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.schema.settings.afterAll.length === 0)
                return;
            for (let i = 0; i < this.schema.settings.afterAll.length; i++) {
                this.callback({ currentTable: '', step: 'afterAll', state: 'RUNNING', currentValue: i, max: this.schema.settings.afterAll.length });
                yield this.dbConnector.executeRawQuery(this.schema.settings.afterAll[i]);
            }
            this.callback({ currentTable: '', step: 'afterAll', state: 'DONE', currentValue: this.schema.settings.afterAll.length, max: this.schema.settings.afterAll.length });
        });
    }
    before(table) {
        return __awaiter(this, void 0, void 0, function* () {
            if (table.before.length === 0)
                return;
            for (let i = 0; i < table.before.length; i++) {
                this.callback({ currentTable: table.name, state: 'RUNNING', step: 'before', currentValue: i, max: table.before.length });
                yield this.dbConnector.executeRawQuery(table.before[i]);
            }
            this.callback({ currentTable: table.name, step: 'before', state: 'DONE', currentValue: table.after.length, max: table.after.length });
        });
    }
    generateData(table) {
        return __awaiter(this, void 0, void 0, function* () {
            const tableForeignKeyValues = {};
            let previousRunRows = -1;
            let currentNbRows = yield this.dbConnector.countLines(table);
            let maxLines = 0;
            if (table.addLines) {
                maxLines = currentNbRows + table.addLines;
                if (table.maxLines !== undefined)
                    maxLines = Math.min(maxLines, table.maxLines);
            }
            else if (table.maxLines)
                maxLines = table.maxLines;
            let insertedRows = 0;
            let deltaRows = maxLines - currentNbRows;
            try {
                yield this.getForeignKeyValues(table, tableForeignKeyValues, deltaRows);
            }
            catch (ex) {
                this.callback({ currentTable: table.name, step: 'generateData', state: 'RUNNING', currentValue: currentNbRows, max: maxLines, comment: ex.message });
            }
            let currentTableRow = 0;
            this.callback({ currentTable: table.name, step: 'generateData', state: 'RUNNING', currentValue: currentNbRows, max: maxLines });
            TABLE_LOOP: while (currentNbRows < maxLines) {
                previousRunRows = currentNbRows;
                const rows = [];
                const runRows = Math.min(1000, maxLines - currentNbRows);
                BATCH_LOOP: for (let currentBatchRow = 0; currentBatchRow < runRows; currentBatchRow++) {
                    const row = {};
                    for (var c = 0; c < table.columns.length; c++) {
                        const column = table.columns[c];
                        if (column.autoIncrement)
                            continue;
                        if (column.values) {
                            let parsedValues = [];
                            let values = column.values;
                            if (typeof values === 'string') {
                                values = this.schema.settings.values[values];
                            }
                            if (!(values instanceof Array)) {
                                Object.keys(values).forEach((key) => {
                                    let arr = new Array(Math.round(values[key] * 100));
                                    arr = arr.fill(key);
                                    parsedValues = parsedValues.concat(arr);
                                });
                            }
                            else {
                                parsedValues = values;
                            }
                            row[column.name] = this.random.pick(parsedValues);
                            continue;
                        }
                        if (column.foreignKey) {
                            const foreignKeys = tableForeignKeyValues[`${column.name}_${column.foreignKey.table}_${column.foreignKey.column}`];
                            if (currentTableRow >= foreignKeys.length && !column.nullable && column.unique) {
                                this.callback({ currentTable: table.name, step: 'generateData', state: 'RUNNING', currentValue: currentNbRows, max: maxLines, comment: colors_1.default.red(`${table.name}: Not enough values available for foreign key ${table.name}.${column.name}`) });
                                break BATCH_LOOP;
                            }
                            if (currentTableRow >= foreignKeys.length && column.unique && column.nullable)
                                continue;
                            row[column.name] = foreignKeys[currentTableRow % foreignKeys.length];
                            continue;
                        }
                        switch (column.generator) {
                            case 'set':
                            case 'bit':
                                row[column.name] = this.random.integer(0, Math.pow(2, column.max));
                                break;
                            case 'bool':
                            case 'boolean':
                                row[column.name] = this.random.bool();
                                break;
                            case 'smallint':
                            case 'mediumint':
                            case 'tinyint':
                            case 'int':
                            case 'integer':
                            case 'bigint':
                                row[column.name] = this.random.integer(column.min, column.max);
                                break;
                            case 'decimal':
                            case 'dec':
                            case 'float':
                            case 'double':
                                row[column.name] = this.random.real(column.min, column.max);
                                break;
                            case 'date':
                            case 'datetime':
                            case 'timestamp':
                                const min = column.min ? new Date(column.min) : new Date('01-01-1970');
                                const max = column.max ? new Date(column.max) : new Date();
                                row[column.name] = this.random.date(min, max);
                                break;
                            case 'time':
                                const hours = this.random.integer(-838, +838);
                                const minutes = this.random.integer(-0, +59);
                                const seconds = this.random.integer(-0, +59);
                                row[column.name] = `${hours}:${minutes}:${seconds}`;
                                break;
                            case 'year':
                                row[column.name] = this.random.integer(column.min, column.max);
                                break;
                            case 'varchar':
                            case 'char':
                            case 'binary':
                            case 'varbinary':
                            case 'tinytext':
                            case 'text':
                            case 'mediumtext':
                            case 'longtext':
                            case 'tinyblob':
                            case 'blob':
                            case 'mediumblob': // 16777215
                            case 'longblob': // 4,294,967,295
                                if (column.max >= 36 && column.unique) {
                                    row[column.name] = this.random.uuid4();
                                }
                                else {
                                    row[column.name] = this.random.string(this.random.integer(column.min, column.max));
                                }
                                break;
                            case 'enum':
                                row[column.name] = Math.floor(this.random.realZeroToOneExclusive() * (column.max)) + 1;
                                break;
                        }
                        if (column.nullable && this.random.realZeroToOneExclusive() <= 0.1)
                            row[column.name] = null;
                    }
                    rows.push(row);
                    currentTableRow++;
                }
                insertedRows = yield this.dbConnector.insert(table.name, rows);
                currentNbRows += insertedRows;
                if (previousRunRows === currentNbRows) {
                    break TABLE_LOOP;
                }
                this.callback({ currentTable: table.name, step: 'generateData', state: 'RUNNING', currentValue: currentNbRows, max: maxLines });
                if (this.continue) {
                    this.continue = false;
                    break TABLE_LOOP;
                }
            }
            this.callback({ currentTable: table.name, step: 'generateData', state: 'DONE', currentValue: currentNbRows, max: maxLines });
            return insertedRows;
        });
    }
    after(table) {
        return __awaiter(this, void 0, void 0, function* () {
            if (table.after.length === 0)
                return;
            for (let i = 0; i < table.after.length; i++) {
                this.callback({ currentTable: table.name, step: 'after', state: 'RUNNING', currentValue: i, max: table.after.length });
                yield this.dbConnector.executeRawQuery(table.after[i]);
            }
            this.callback({ currentTable: table.name, step: 'after', state: 'DONE', currentValue: table.after.length, max: table.after.length });
        });
    }
}
exports.Generator = Generator;
